@@ -3,7 +3,7 @@
 # =========================================================================
 #   Program:   S1Processor
 #
-#   Copyright 2017-2021 (c) CESBIO. All rights reserved.
+#   Copyright 2017-2021 (c) CNES. All rights reserved.
 #
 #   This file is part of S1Tiling project
 #       https://gitlab.orfeo-toolbox.org/s1-tiling/s1tiling
@@ -242,10 +242,16 @@ def process_one_tile(
     logger.debug('Summary of tasks related to S1 -> S2 transformations of %s', tile_name)
     results = []
     if debug_otb:
-        for product, how in reversed(dsk):
+        tasks = list(Utils.tsort(dsk, dsk.keys(),
+            lambda dask_task_data : [] if isinstance(dask_task_data, FirstStep) else dask_task_data[2])
+            )
+        logger.debug('%s tasks', len(tasks))
+        for product in reversed(tasks):
+            how = dsk[product]
             logger.debug('- task: %s <-- %s', product, how)
         logger.info('Executing tasks one after the other for %s (debugging OTB)', tile_name)
-        for product, how in reversed(dsk):
+        for product in reversed(tasks):
+            how = dsk[product]
             logger.info('- execute: %s <-- %s', product, how)
             if not issubclass(type(how), FirstStep):
                 results += [how[0](*list(how)[1:])]
@@ -260,7 +266,7 @@ def process_one_tile(
                     filename='tasks-%s-%s.svg' % (tile_idx + 1, tile_name))
         logger.info('Start S1 -> S2 transformations for %s', tile_name)
         nb_tries = 2
-        for run in range(1, nb_tries+1):
+        for run_attemp in range(1, nb_tries+1):
             try:
                 results = client.get(dsk, required_products)
                 return results
@@ -273,14 +279,20 @@ def process_one_tile(
                 # but then, how do we clean up futures and all??
                 client.restart()
                 # Update the list of remaining tasks
-                if run < nb_tries:
+                if run_attemp < nb_tries:
                     dsk, required_products = pipelines.generate_tasks(tile_name, intersect_raster_list,
                             debug_otb=debug_otb, dryrun=dryrun, do_watch_ram=do_watch_ram)
                 else:
                     raise
 
 
-def s1_process(searched_items_per_page, dryrun, debug_otb, watch_ram, debug_tasks, cache_before_ortho, config_filename):
+def s1_process(config_opt,
+               searched_items_per_page=20,
+               dryrun=False,
+               debug_otb=False,
+               watch_ram=False,
+               debug_tasks=False,
+               cache_before_ortho=False):
     """
       On demand Ortho-rectification of Sentinel-1 data on Sentinel-2 grid.
 
@@ -293,7 +305,12 @@ def s1_process(searched_items_per_page, dryrun, debug_otb, watch_ram, debug_task
 
       Parameters have to be set by the user in the S1Processor.cfg file
     """
-    config = Configuration(config_filename)
+    # The config_opt can be either the configuration filename or an already initialized configuration object
+    if isinstance(config_opt, str):
+        config = Configuration(config_opt)
+    else:
+        config = config_opt
+
     os.environ["ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS"] = str(config.OTBThreads)
     global logger
     logger = logging.getLogger('s1tiling')
@@ -342,43 +359,49 @@ def s1_process(searched_items_per_page, dryrun, debug_otb, watch_ram, debug_task
 
         # filtering_processor = S1FilteringProcessor.S1FilteringProcessor(config)
 
-        if not debug_otb:
-            clean_logs(config.log_config, config.nb_procs)
-            cluster = LocalCluster(threads_per_worker=1, processes=True, n_workers=config.nb_procs, silence_logs=False)
-            client = Client(cluster)
-            client.register_worker_callbacks(lambda dask_worker: setup_worker_logs(config.log_config, dask_worker))
-        else:
-            client = None
+        try:
+            if not debug_otb:
+                clean_logs(config.log_config, config.nb_procs)
+                cluster = LocalCluster(threads_per_worker=1, processes=True, n_workers=config.nb_procs, silence_logs=False)
+                client = Client(cluster)
+                client.register_worker_callbacks(lambda dask_worker: setup_worker_logs(config.log_config, dask_worker))
+            else:
+                client = None
 
-        log_level = lambda res: logging.INFO if bool(res) else logging.WARNING
-        results = []
-        for idx, tile_it in enumerate(tiles_to_process_checked):
-            with Utils.ExecutionTimer("Processing of tile " + tile_it, True):
-                res = process_one_tile(
-                        tile_it, idx, len(tiles_to_process_checked),
-                        s1_file_manager, pipelines, client,
-                        searched_items_per_page=searched_items_per_page,
-                        debug_otb=debug_otb, dryrun=dryrun, do_watch_ram=watch_ram, debug_tasks=debug_tasks)
-                results += res
+            log_level = lambda res: logging.INFO if bool(res) else logging.WARNING
+            results = []
+            for idx, tile_it in enumerate(tiles_to_process_checked):
+                with Utils.ExecutionTimer("Processing of tile " + tile_it, True):
+                    res = process_one_tile(
+                            tile_it, idx, len(tiles_to_process_checked),
+                            s1_file_manager, pipelines, client,
+                            searched_items_per_page=searched_items_per_page,
+                            debug_otb=debug_otb, dryrun=dryrun, do_watch_ram=watch_ram, debug_tasks=debug_tasks)
+                    results += res
 
-        nb_error_detected = 0
-        for res in results:
-            if not bool(res):
-                nb_error_detected += 1
-
-        if nb_error_detected > 0:
-            logger.warning('Execution report: %s errors detected', nb_error_detected)
-        else:
-            logger.info('Execution report: no error detected')
-
-        if results:
+            nb_error_detected = 0
             for res in results:
-                logger.log(log_level(res), ' - %s', res)
-        else:
-            logger.info(' -> Nothing has been executed')
+                if not bool(res):
+                    nb_error_detected += 1
+
+            if nb_error_detected > 0:
+                logger.warning('Execution report: %s errors detected', nb_error_detected)
+            else:
+                logger.info('Execution report: no error detected')
+
+            if results:
+                for res in results:
+                    logger.log(log_level(res), ' - %s', res)
+            else:
+                logger.info(' -> Nothing has been executed')
+
+            return nb_error_detected
+
+        finally:  # Make sure Dask objects are ALWAYS released
+            if client:
+                client.close()
+                cluster.close()
 
-        if nb_error_detected > 0:
-            sys.exit(exits.TASK_FAILED)
 
 @click.command(context_settings=dict(help_option_names=["-h", "--help"]))
 @click.version_option()
@@ -415,9 +438,18 @@ def run( searched_items_per_page, dryrun, debug_otb, watch_ram,
          debug_tasks, cache_before_ortho, config_filename):
     """
     This function is used as entry point to create console scripts with setuptools.
+
+    Returns the number of tasks that could not be processed.
     """
-    s1_process( searched_items_per_page, dryrun, debug_otb, watch_ram,
-                debug_tasks, cache_before_ortho, config_filename)
+    nb_error_detected = s1_process( config_filename,
+                searched_items_per_page=searched_items_per_page,
+                dryrun=dryrun,
+                debug_otb=debug_otb,
+                watch_ram=watch_ram,
+                debug_tasks=debug_tasks,
+                cache_before_ortho=cache_before_ortho)
+    if nb_error_detected > 0:
+        sys.exit(exits.TASK_FAILED)
 
 if __name__ == '__main__':  # Required for Dask: https://github.com/dask/distributed/issues/2422
     run()
